#include <linux/config.h>
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/pci.h>
#include <asm-i386/param.h>

#include <asm/uaccess.h>
#include "tdsvar.h"
#include "tdsio.h"

#ifndef CONFIG_PCI
#error "Requires PCI support"
#endif


#ifdef CONFIG_DEVFS_FS
#error "devfs is not supported (yet)"
#endif


MODULE_DESCRIPTION("Driver for Nitsuki Time Sync Data Sampler");
MODULE_SUPPORTED_DEVICE(NITSUKI_TDS_DRIVER_NAME);
MODULE_AUTHOR("Kimura Moritaka [mkimura@nict.go.jp]");
MODULE_LICENSE("Proprietary.  Send bug reports to mkimura@nict.go.jp");

static char tds_ver[] = "$Id: tds.c,v 2.01 2005/06/28, Modification version writen by Kimura Moritaka. $";


#define TDS_PRINT(level, fmt, args...)	\
	printk(level "%s: %s: " fmt "\n" , \
		tds_driver_name, __FUNCTION__, ## args)

#define TDS_SC_PRINT(level, fmt, args...)	\
	printk(level "%s: %s: " fmt "\n" , \
		sc->dev_name, __FUNCTION__, ## args)

#define TDS_ASSERT(cond)	\
	do {								\
		if (!(cond)) {					\
			printk (KERN_EMERG				\
				"Assertion failure in %s at %s:%d: \"%s\"\n", \
				__FUNCTION__, __FILE__, __LINE__, # cond); \
			BUG();						\
		}							\
	} while (0)


#ifndef MIN
#define MIN(a, b)	(((a) < (b)) ? (a) : (b))
#endif

#ifndef BITSPERBYTE
#define BITSPERBYTE	8
#endif

#define TDS_INTR_SET_PIO_MODE \
	(TDS_INTR_DATA_EXIST | TDS_INTR_DATA_FULL)

#define TDS_INTR_SET_DMA_MODE \
	(TDS_INTR_DMA_DONE | TDS_INTR_DATA_FULL)


enum tds_bool {
	tds_true = 1,
	tds_false = 0,
};


static int omit_timeset_check = 0;

MODULE_PARM(omit_timeset_check, "i");
MODULE_PARM_DESC(omit_timeset_check,
		 "If not 0, omit checking whether time is set (or sync)");


enum tds_debug_msg_level {
	tds_debug_none = 0,
	tds_debug_minimum = 1,
	tds_debug_medium = 2,
	tds_debug_most = 3,
};


#ifdef _TDS_DEBUG
#define TDS_INFINITE	300
#define	TDS_INLINE	/* none */
/*static int tds_debug = tds_debug_medium;*/
static int tds_debug = tds_debug_most;
#define TDS_SC_DEBUG(dlevel, fmt, args...)			\
	do {							\
		if (tds_debug >= (dlevel)) {			\
			TDS_SC_PRINT(KERN_DEBUG, fmt, ## args);	\
		}						\
	} while (0)
#define TDS_DEBUG(dlevel, fmt, args...)				\
	do {							\
		if (tds_debug >= (dlevel)) {			\
			TDS_PRINT(KERN_DEBUG, fmt, ## args);	\
		}						\
	} while (0)
#else	/* _TDS_DEBUG */
#define TDS_INFINITE	MAX_SCHEDULE_TIMEOUT
#define	TDS_INLINE	__inline__
static int tds_debug = tds_debug_none;
#define TDS_SC_DEBUG(dlevel, fmt, args...)	/* none */
#define TDS_DEBUG(dlevel, fmt, args...)		/* none */
#endif	/* _TDS_DEBUG */

MODULE_PARM(tds_debug, "i");
MODULE_PARM_DESC(tds_debug, "debug message level control");


/* driver name */
static char tds_driver_name[] = NITSUKI_TDS_DRIVER_NAME;

static struct tds_softc *tds_info;
static int dev_count = 0;

static struct pci_device_id tds_pci_tbl[] __devinitdata = {
	{
		vendor:		NITSUKI_VENDERID,
		device:		NITSUKI_TIME_SYNC_DATA_SAMPLER,
		subvendor:	PCI_ANY_ID,
		subdevice:	PCI_ANY_ID,
	},
	{ 0, },	/* terminator */
};
MODULE_DEVICE_TABLE(pci, tds_pci_tbl);


static struct tds_chip_info {
	char *name;
	unsigned short vendor;
	unsigned short device;
} tds_info_tbl[] __devinitdata = {
	{ "Nitsuki Time Sync Data Sampler",
	  NITSUKI_VENDERID, NITSUKI_TIME_SYNC_DATA_SAMPLER },
	{ NULL, 0, 0 }	/* terminator */
};


static int __devinit tds_init_one(struct pci_dev *pdev,
				  const struct pci_device_id *ent);
static void __devexit tds_remove_one(struct pci_dev *pdev);

static struct pci_driver tds_pci_driver = {
	name:		tds_driver_name,
	id_table:	tds_pci_tbl,
	probe:		tds_init_one,
	remove:		__devexit_p(tds_remove_one),
};


static void tds_irq_handler(int irq, void *dev_id, struct pt_regs *ignored);


static int tds_open(struct inode *inode, struct file *filp);
static int tds_release(struct inode *inode, struct file *filp);
static ssize_t tds_read(struct file *filp, char *buf,
			size_t count, loff_t *ignored);
static ssize_t tds_write(struct file *filp, const char *buf,
			 size_t count, loff_t *ignored);
static int tds_ioctl(struct inode *inode, struct file *filp,
		     unsigned int cmd, unsigned long arg);

static struct file_operations tds_fops = {
	owner:		THIS_MODULE,
	open:		tds_open,
	release:	tds_release,
	read:		tds_read,
	write:		tds_write,
	ioctl:		tds_ioctl,
};


static int __init tds_init_module(void);
static void __exit tds_cleanup_module(void);

module_init(tds_init_module);
module_exit(tds_cleanup_module);


/* device register handling */
static TDS_INLINE void tdsreg_clear_fifo(struct tds_softc *sc);
static TDS_INLINE void tdsreg_sync_1pps(struct tds_softc *sc);
static TDS_INLINE void tdsreg_sync_time(struct tds_softc *sc);
static TDS_INLINE void tdsreg_set_irq(struct tds_softc *sc, u8 flags);
static TDS_INLINE u8 tdsreg_get_irq(struct tds_softc *sc);
static TDS_INLINE void tdsreg_set_time(struct tds_softc *sc, u32 time);
static TDS_INLINE u32 tdsreg_get_time(struct tds_softc *sc);
static TDS_INLINE u8 tdsreg_get_year(struct tds_softc *sc);
static TDS_INLINE void tdsreg_set_sampling(struct tds_softc *sc, u8 flags);
static TDS_INLINE u8 tdsreg_get_status(struct tds_softc *sc);
static TDS_INLINE void tdsreg_start_sampling(struct tds_softc *sc);
static TDS_INLINE void tdsreg_stop_sampling(struct tds_softc *sc);
static TDS_INLINE u32 tdsreg_get_data(struct tds_softc *sc);
static TDS_INLINE u8 tdsreg_get_input_status(struct tds_softc *sc);
static TDS_INLINE void tdsreg_set_dma_address(struct tds_softc *sc, u32 dstaddr);
static TDS_INLINE void tdsreg_set_dma_datasize(struct tds_softc *sc, u32 dsize);
static TDS_INLINE u32 tdsreg_get_dma_datasize(struct tds_softc *sc);
static TDS_INLINE void tdsreg_start_dma(struct tds_softc *sc);
static TDS_INLINE void tdsreg_stop_dma(struct tds_softc *sc);
static TDS_INLINE u8 tdsreg_get_dma_status(struct tds_softc *sc);
static TDS_INLINE void tdsreg_set_dma_mode(struct tds_softc *sc, u8 mode);
#if 0
static TDS_INLINE u8 tdsreg_get_dma_mode(struct tds_softc *sc);
static TDS_INLINE u32 tdsreg_get_write_counter(struct tds_softc *sc);
static TDS_INLINE u32 tdsreg_get_read_counter(struct tds_softc *sc);
static TDS_INLINE void tdsreg_clear_write_counter(struct tds_softc *sc);
static TDS_INLINE void tdsreg_clear_read_counter(struct tds_softc *sc);
#endif

/* support functions */
static TDS_INLINE void tds_wait_next_1pps(struct tds_softc *sc);
static TDS_INLINE void tds_issue_dma(struct tds_softc *sc);
static TDS_INLINE int tds_check_extension_card(struct tds_softc *sc);
static TDS_INLINE void tds_dma_buffer_init(struct tds_softc *sc);
static TDS_INLINE void tds_calc_dma_chunk_size(struct tds_softc *sc);
static TDS_INLINE int tds_sampling_condition(struct tds_softc *sc, u32 cond);
static TDS_INLINE void tds_dma_buf_free(struct tds_softc *sc);
static TDS_INLINE int tds_dma_buf_alloc(struct tds_softc *sc);
static TDS_INLINE int tds_init(struct tds_softc *sc, int mode);
#if 0
static TDS_INLINE int tds_get_fifo_count(struct tds_softc *sc);
#endif
static TDS_INLINE int tdsio_sampling_start(struct tds_softc *sc);
static TDS_INLINE void tdsio_sampling_stop(struct tds_softc *sc);


/* driver routine */
static int __devinit
tds_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int err = 0;
	struct tds_softc *sc;

	TDS_DEBUG(tds_debug_minimum, "called, dev_count = %d", dev_count);

	if (dev_count >= 1) {
		TDS_PRINT(KERN_ERR, "this driver only support 1 device");
		return (-ENODEV);
	}
	else {
		struct tds_chip_info *info = tds_info_tbl;
		while (info->name != NULL) {
			if ((ent->vendor == info->vendor)
			    && (ent->device == info->device)) {
				TDS_PRINT(KERN_INFO, "%s", info->name);
				break;
			}
			info ++;
		}
		TDS_PRINT(KERN_INFO, "%s", tds_ver);
	}

	if (pci_enable_device(pdev)) {
		err = (-ENXIO);
		goto err_out_none;
	}

	pci_set_master(pdev);

	tds_info = kmalloc(sizeof(*tds_info), GFP_KERNEL);
	if (tds_info == NULL) {
		TDS_PRINT(KERN_ERR, "kmalloc failed");
		err = (-ENOMEM);
		goto err_out_pci;
	}
	sc = tds_info;
	memset(sc, 0, sizeof(*sc));

	pci_set_drvdata(pdev, sc);

	snprintf(sc->dev_name, sizeof(sc->dev_name),
		 "%s%d", tds_driver_name, dev_count);
	dev_count ++;

	sc->pdev = pdev;
	sc->iobase = pci_resource_start(pdev, 0);
	sc->res_io = request_region(sc->iobase, pci_resource_len(pdev, 0),
				    sc->dev_name);
	if (sc->res_io == NULL) {
		TDS_SC_PRINT(KERN_ERR, "cannot reserve I/O ports");
		err = (-ENODEV);
		goto err_out_softc;
	}
	else {
		TDS_SC_DEBUG(tds_debug_medium, "IOBASE = 0x%x", sc->iobase);
	}

	sc->irq = pdev->irq;
	TDS_SC_DEBUG(tds_debug_medium, "IRQ = %d", sc->irq);
	err = request_irq(sc->irq, tds_irq_handler, SA_SHIRQ,
			  sc->dev_name, (void *)sc);
	if (err != 0) {
		TDS_SC_PRINT(KERN_ERR, "request_irq failed");
		goto err_out_resio;
	}
	
	//pci_write_config_byte(pdev,0x0d,0x80); //append M.K.
	pci_write_config_byte(pdev,0x0d,0x20); //append M.K.


#if 0	/* this might be verbose */
	{
		u16 data;
		err = pci_read_config_word(pdev, PCI_COMMAND, &data);
		if (err != 0) {
			goto err_out_irq;
		}
		TDS_SC_DEBUG(tds_debug_most, "Command Reg: %04xh", data);
		data |= (PCI_COMMAND_IO | PCI_COMMAND_MASTER);
		err = pci_write_config_word(pdev, PCI_COMMAND, data);
		if (err != 0) {
			goto err_out_irq;
		}
		err = pci_read_config_word(pdev, PCI_COMMAND, &data);
		if (err != 0) {
			goto err_out_irq;
		}
		TDS_SC_DEBUG(tds_debug_most, "Command Reg: %04xh", data);
	}
#endif

	sc->intr_flag = 0;
	tdsreg_set_irq(sc, sc->intr_flag);
	tdsreg_stop_sampling(sc);

	err = tds_init(sc, TDS_READ_DEFAULT);
	if (err != 0) {
		TDS_SC_PRINT(KERN_ERR, "tds_init failed");
		goto err_out_irq;
	}

	return 0;

 err_out_irq:
	free_irq(sc->irq, (void *)sc);

 err_out_resio:
	release_region(pci_resource_start(pdev, 0), pci_resource_len(pdev, 0));

 err_out_softc:
	tds_dma_buf_free(sc);
	kfree(sc);

 err_out_pci:
	pci_set_drvdata(pdev, NULL);
	pci_disable_device(pdev);
	dev_count --;

 err_out_none:
	return err;
}

static void __devexit
tds_remove_one(struct pci_dev *pdev)
{
	struct tds_softc *sc = (struct tds_softc *)pci_get_drvdata(pdev);
	
	TDS_SC_DEBUG(tds_debug_minimum, "called");
	
	if (sc != NULL) {
		free_irq(sc->irq, (void *)sc);
		release_region(pci_resource_start(pdev, 0),
			       pci_resource_len(pdev, 0));
		tds_dma_buf_free(sc);
		kfree(sc);
	}
	pci_set_drvdata(pdev, NULL);
	pci_disable_device(pdev);
	dev_count --;
}

int read_fifo_counter(struct tds_softc *sc)
{
	return inl(sc->iobase +TDS_REG_READ_READ_COUNTER);
}

static void tds_irq_handler(int irq, void *dev_id, struct pt_regs *ignored)
{
	struct tds_softc *sc = (struct tds_softc *)dev_id;
	u8 intr;
	struct timeval time;
	if (irq != sc->irq) {	/* might be verbose */
		return;
	}
	
	intr = tdsreg_get_irq(sc) & sc->intr_flag;
	
	if (intr == 0) {	/* nothing to do */
		return;
	}
	
	do_gettimeofday(&time);
	if (intr & TDS_INTR_DATA_EXIST) {
		wake_up_interruptible(&sc->wait_sync_data);
	}
	
	if (intr & TDS_INTR_1PPS) {
		wake_up_interruptible(&sc->wait_sync_1pps);
	}
	if (intr & TDS_INTR_DATA_HALF) {
		TDS_SC_DEBUG(tds_debug_medium, "FIFO half at %ld.%ld",
			     time.tv_sec, time.tv_usec);
		wake_up_interruptible(&sc->wait_fifo_overhalf);
	}
	
	if (intr & TDS_INTR_DATA_FULL) {
		TDS_SC_DEBUG(tds_debug_medium, "FIFO full at %ld.%ld",
			     time.tv_sec, time.tv_usec);
		sc->fifo_overflow_count ++;
		tdsreg_clear_fifo(sc);
		wake_up_interruptible(&sc->wait_fifo_overflow);
	}	
	if (intr & TDS_INTR_DMA_DONE) {
		unsigned long irq_flag;
		unsigned int *adr1,*adr2;
		int current_fifo,read_size,dma_size;
		static int pre_fifo=-1;
		spin_lock_irqsave(&sc->lock, irq_flag);	
		dma_size=tdsreg_get_dma_datasize(sc);
		if(dma_size==0) {pre_fifo=-1;goto skip;}
		current_fifo=read_fifo_counter(sc);
		if(pre_fifo>=0) read_size=current_fifo-pre_fifo;else read_size=current_fifo-1;
		while(read_size<0) read_size+=0x40000000;
		if(read_size!=dma_size) printk("TDS-DRIVER Warning:read_size=%08x,dma_size=%08x\n",read_size,dma_size);
		
		pre_fifo=current_fifo;
		if(read_size>=2*TDS_DMA_MAX_CHUNK) read_size=2*TDS_DMA_MAX_CHUNK;
		adr1=(unsigned int *)sc->dmabuf[TDS_DMA_NSEG-1  ].virtaddr;
		adr2=(unsigned int *)sc->dmabuf[sc->dmabuf_write].virtaddr;
		//printk("DMA buf[%02d][%08x-%08x] %d.%06d\n",sc->dmabuf_write,0,read_size,(unsigned int)time.tv_sec,(unsigned int)time.tv_usec);
		int s;for(s=0;s<read_size;s++) adr2[s]=adr1[s];
		sc->dmabuf[sc->dmabuf_write].write_offset=4*read_size;
		sc->dmabuf_write=(sc->dmabuf_write + 1) % (TDS_DMA_NSEG-1);
		sc->dmabuf[sc->dmabuf_write].write_offset=0;
	skip:
		spin_unlock_irqrestore(&sc->lock, irq_flag);
		wake_up_interruptible(&sc->wait_buffer_ready);
	}
}

static int
tds_open(struct inode *inode, struct file *filp)
{
	struct tds_softc *sc;

	TDS_DEBUG(tds_debug_minimum, "called");

	if (filp->private_data == NULL) {
		sc = tds_info;
		filp->private_data = sc;
	}
	else {
		sc = (struct tds_softc *)filp->private_data;
	}
	
	if (sc->is_open == tds_true) {
		return (-EBUSY);
	}

	sc->fifo_overflow_count = 0;
	sc->is_open = tds_true;

	return 0;
}


static int
tds_release(struct inode *inode, struct file *filp)
{
	struct tds_softc *sc = (struct tds_softc *)filp->private_data;

	TDS_SC_DEBUG(tds_debug_minimum, "called");

	if (sc->sampling_start == tds_true) {
		tdsio_sampling_stop(sc);
	}

	sc->is_open = tds_false;

	return 0;
}


static ssize_t
tds_read(struct file *filp, char *buf, size_t count, loff_t *ignored)
{
	struct tds_softc *sc = (struct tds_softc *)filp->private_data;
	ssize_t ret = 0;
	int *idx = &sc->dmabuf_read;
	int data_count = 0;
	/*
	if (down_interruptible(&sc->sem)) {
		TDS_SC_PRINT(KERN_ERR, "can't get semaphore");
		return (-ERESTARTSYS);
	}
	*/
	while ((count - data_count) > 0) {
		tds_dma_buf_t *dmabuf = &(sc->dmabuf[*idx]);
		size_t size;
		unsigned int addr;
		while (sc->dmabuf_read == sc->dmabuf_write) {current->state=TASK_INTERRUPTIBLE;schedule_timeout(HZ/100);}
		size = MIN(count - data_count,dmabuf->write_offset-dmabuf->read_offset);
		addr = dmabuf->virtaddr+dmabuf->read_offset;
		ret = copy_to_user((void *)(buf + data_count),(void *)addr, size);
		//printk("Send to APP buf[%02d][0x%08x,%08d] size=%d\n",sc->dmabuf_read,dmabuf->read_offset,dmabuf->read_offset+size,size);
		data_count += size;dmabuf->read_offset+=size;
		if(dmabuf->write_offset==dmabuf->read_offset) {
			*idx = (*idx + 1) % (TDS_DMA_NSEG-1);
			dmabuf = &(sc->dmabuf[*idx]);
			dmabuf->read_offset=0;
		}
	}
	if (ret == 0) {
		ret = data_count;
	}
	else {
		ret = (-EFAULT);
	}
	//up(&sc->sem);
	return ret;
}


static ssize_t
tds_write(struct file *filp, const char *buf, size_t count, loff_t *ignored)
{
	struct tds_softc *sc = (struct tds_softc *)filp->private_data;

	TDS_SC_DEBUG(tds_debug_minimum, "called "
		     "(filp = 0x%p, buf = 0x%p, count = %d, offp = 0x%p)",
		     filp, buf, count, ignored);

	TDS_SC_PRINT(KERN_ERR, "write operation not supported");
	sc->error = TDS_ERROR_NOT_SUPPORTED;

	return (-EINVAL);
}


static int
tds_ioctl(struct inode *inode, struct file *filp,
	  unsigned int cmd, unsigned long arg)
{
	struct tds_softc *sc = (struct tds_softc *)filp->private_data;
	int rv = 0;

	TDS_SC_DEBUG(tds_debug_minimum, "called, cmd = 0x%08x", (int)cmd);

	switch (cmd) {
	case TDSIO_INIT:
		if (sc->sampling_start == tds_true) {
			TDS_SC_PRINT(KERN_ERR,
				     "can't initialize "
				     "while sampling in progress");
			sc->error = TDS_ERROR_ALREADY_STARTED;
			return (-EBUSY);
		}
		tds_init(sc, TDS_READ_DEFAULT);
		return 0;
	case TDSIO_BUFFER_CLEAR:
		if (sc->sampling_start == tds_true) {
			TDS_SC_PRINT(KERN_ERR,
				     "can't clear FIFO "
				     "while sampling in progress");
			sc->error = TDS_ERROR_ALREADY_STARTED;
			return (-EBUSY);
		}
		tdsreg_clear_fifo(sc);
		TDS_SC_DEBUG(tds_debug_medium, "cleared device FIFO");
		return 0;
	case TDSIO_SYNC_1PPS: {
		int count;
		for (count = 0; count < 2; count ++) {
			u8 status = tdsreg_get_status(sc);
			if (status & TDS_STAT_1PPS_INPUT) {
				tdsreg_sync_1pps(sc);
				sc->pps_sync_done = tds_true;
				TDS_SC_DEBUG(tds_debug_medium,
					     "synchronized to 1pps");
				return 0;
			}
			/* XXX: one more try */
			tds_wait_next_1pps(sc);
		}

		TDS_SC_PRINT(KERN_ERR, "1pps input not detected");
		sc->error = TDS_ERROR_1PPS_NOT_DETECT;
		return (-EIO);
	}
	case TDSIO_SYNC_TIME: {
		int count;
		for (count = 0; count < 2; count ++) {
			u8 status;
			status = tdsreg_get_status(sc);
			if (status & TDS_STAT_TIME_INPUT) {
				tdsreg_sync_time(sc);
				tds_wait_next_1pps(sc);
				sc->time_sync_done = tds_true;
				TDS_SC_DEBUG(tds_debug_medium,
					     "synchronized to time");
				return 0;
			}
			/* XXX: one more try */
			tds_wait_next_1pps(sc);
		}
		if (sc->time_sync_done == tds_false) {
			TDS_SC_PRINT(KERN_ERR, "time input not detected");
			sc->error = TDS_ERROR_TIME_NOT_DETECT;
			rv = (-EIO);
		}
		break;
	}
	case TDSIO_SET_INTERRUPT:	/* OBSOLETE */
	case TDSIO_GET_INTERRUPT:	/* OBSOLETE */
		break;	/* nothing to do */
	case TDSIO_SET_TIME: {
		u32 *time = (u32 *)arg;
		tdsreg_set_time(sc, *time);
		sc->time_set_done = tds_true;
		TDS_SC_DEBUG(tds_debug_medium, "set time 0x%08x", *time);
		break;
	}
	case TDSIO_GET_TIME: {
		u32 *time = (u32 *)arg;
		*time = tdsreg_get_time(sc);
		break;
	}
	case TDSIO_GET_YEAR: {
		u8 *year = (u8 *)arg;
		*year = tdsreg_get_year(sc);
		break;
	}
	case TDSIO_SET_SAMPLING: {
		u32 *flags = (u32 *)arg;

		rv = tds_sampling_condition(sc, *flags);
		if (rv != 0) {
			break;
		}
		tdsreg_set_sampling(sc, *flags);
		sc->sampling_flag = *flags;
		sc->sampling_done = tds_true;
		break;
	}
	case TDSIO_GET_SAMPLING: {
		u32 *flags = (u32 *)arg;
		if (sc->sampling_done == tds_false) {
			TDS_SC_PRINT(KERN_ERR, "not set sampling yet");
			sc->error = TDS_ERROR_NOT_SET_SAMPLING;
			rv = (-ENXIO);
		}
		else {
			*flags = sc->sampling_flag;
		}
		break;
	}
	case TDSIO_GET_STATUS: {
		u32 *stat = (u32 *)arg;
		u8 tmp;
		tmp = tdsreg_get_status(sc);
		*stat = (u32) tmp;
		break;
	}
	case TDSIO_SAMPLING_START:
		if (sc->sampling_start == tds_false) {
			rv = tdsio_sampling_start(sc);
		}
		else {
			TDS_SC_PRINT(KERN_ERR, "sampling already started");
			sc->error = TDS_ERROR_ALREADY_STARTED;
			rv = (-EBUSY);
		}
		break;
	case TDSIO_SAMPLING_STOP:
		if (sc->sampling_start == tds_true) {
			tdsio_sampling_stop(sc);
		}
		else {
			TDS_SC_PRINT(KERN_ERR, "sampling not started yet");
			sc->error = TDS_ERROR_NOT_STARTED;
			rv = (-ENXIO);
		}
		break;
	case TDSIO_GET_ERROR: {
		int *error = (int *)arg;
		*error = sc->error;
		break;
	}
	case TDSIO_SET_VERBOSE: {
		int *verb = (int *)arg;
		sc->verbose = *verb;
		break;
	}
	case TDSIO_SET_MODE: {	/* XXX: obsolete */
		/* nothing to do */
		break;
	}
	case TDSIO_GET_MODE: {	/* XXX: obsolete */
		/* nothing to do */
		break;
	}
	case TDSIO_SET_ALLOW_OVERFLOW: {
		int *num = (int *)arg;
		sc->max_fifo_overflow = *num;
		break;
	}
	case TDSIO_GET_OVERFLOW_COUNT: {
		int *num = (int *)arg;
		*num = sc->fifo_overflow_count;
		break;
	}
	case TDSIO_SET_READ_MODE: {
		if (sc->sampling_start == tds_false) {
			int *mode = (int *)arg;
			switch (*mode) {
			case TDS_READ_USE_PIO:
				sc->read_mode = *mode;
				tds_dma_buf_free(sc);
				break;
			case TDS_READ_USE_DMA:
				sc->read_mode = *mode;
				rv = tds_dma_buf_alloc(sc);
				break;
			default:
				TDS_SC_PRINT(KERN_ERR,
					     "unknown read mode (%08x)",
					     (int)*mode);
				sc->error = TDS_ERROR_NOT_SUPPORTED;
				rv = (-ENOTTY);
				break;
			}
		}
		else {
			TDS_SC_PRINT(KERN_ERR,
				     "can't set read mode "
				     "while sampling in progress");
			sc->error = TDS_ERROR_READ_BUSY;
			rv = (-EBUSY);
		}
		break;
	}
	case TDSIO_GET_READ_MODE: {
		int *mode = (int *)arg;
		*mode = sc->read_mode;
		break;
	}
	case TDSIO_GET_INPUT_STATUS: {
		u8 input;
		int *status = (int *)arg;
		*status = 0;
		input = tdsreg_get_input_status(sc);
		*status += input;
		break;
	}
	case TDSIO_GET_DMA_STATUS: {
		u8 val;
		int *status = (int *)arg;
		*status = 0;
		val = tdsreg_get_dma_status(sc);
		*status += val;
		break;
	}
	default:
		TDS_SC_PRINT(KERN_ERR,
			     "unknown ioctl() command (%08x)", (int)cmd);
		sc->error = TDS_ERROR_NOT_SUPPORTED;
		rv = (-ENOTTY);
		break;
	}

	return rv;
}


static int __init
tds_init_module(void)
{
	int ret = 0;

	TDS_DEBUG(tds_debug_minimum, "called, tds_debug = %d", tds_debug);

	ret = pci_module_init(&tds_pci_driver);
	if (ret != 0) {
		TDS_PRINT(KERN_DEBUG, "pci_module_init failed");
		return ret;
	}

	ret = register_chrdev(TDS_MAJOR, tds_driver_name, &tds_fops);
	if (ret != 0) {
		TDS_PRINT(KERN_ERR, "register_chrdev failed");
		goto pci_unreg;
	}

	EXPORT_NO_SYMBOLS;

	return ret;

 pci_unreg:
	pci_unregister_driver(&tds_pci_driver);

	return ret;
	
}


static void __exit
tds_cleanup_module(void)
{
	TDS_DEBUG(tds_debug_minimum, "called");

	unregister_chrdev(TDS_MAJOR, tds_driver_name);

	pci_unregister_driver(&tds_pci_driver);
}


/* device register handling */

#define	TDS_IOPORT(off)	((u32) (sc->iobase + off))

/*
 * NOTICE:
 *	argument order for 'outb' or 'outl' is different from BSD's
 */
#define TDS_INB(port)		inb(port)
#define TDS_OUTB(port, val)	outb(val, port)
#define TDS_INL(port)		inl(port)
#define TDS_OUTL(port, val)	outl(val, port)


static u8 tds_dbuf_clr_seq[] = { 3, 2, 0, };

static TDS_INLINE void
tdsreg_clear_fifo(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_FIFO_CLEAR);
	int num = sizeof(tds_dbuf_clr_seq) / sizeof(u8);
	int i;
	for (i = 0; i < num; i ++) {
		TDS_OUTB(reg, tds_dbuf_clr_seq[i]);
	}
}


static TDS_INLINE void
tdsreg_sync_1pps(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_SYNCRONIZE);
	TDS_OUTB(reg, TDS_1PPS_SYNC);
	TDS_OUTB(reg, 0);
}


static TDS_INLINE void
tdsreg_sync_time(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_SYNCRONIZE);
	TDS_OUTB(reg, TDS_TIME_SYNC);
}


static TDS_INLINE void
tdsreg_set_irq(struct tds_softc *sc, u8 flags)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_INTERRUPT_SET);
	TDS_SC_DEBUG(tds_debug_medium, "Set Interrupt (0x%02x)", flags);
	TDS_OUTB(reg, flags);
}


static TDS_INLINE u8
tdsreg_get_irq(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_INTERRUPT);
	u8 val = TDS_INB(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_time(struct tds_softc *sc, u32 time)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_TIME_SET);
	TDS_OUTL(reg, time);
}


static TDS_INLINE u32
tdsreg_get_time(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_TIME);
	u32 val = TDS_INL(reg);
	return val;
}


static TDS_INLINE u8
tdsreg_get_year(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_YEAR);
	u8 val = TDS_INB(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_sampling(struct tds_softc *sc, u8 flags)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_SET);
	TDS_OUTB(reg, flags);
}


static TDS_INLINE u8
tdsreg_get_status(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_STATUS);
	u8 val = TDS_INB(reg);
	TDS_SC_DEBUG(tds_debug_medium, "status = 0x%02x", val);
	return val;
}


static TDS_INLINE void
tdsreg_start_sampling(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_START);
	
	//  ************************************************ //append  M.K.
	int i;
	for (i = 0; i < TDS_DMA_NSEG; i ++) {
		tds_dma_buf_t *buf = &sc->dmabuf[i];
		buf->write_offset = 0;
		buf->read_offset = 0;
		buf->flags = 0;
	}
	sc->dmabuf_write = 0;
	sc->dmabuf_read = 0;
	tdsreg_clear_fifo(sc);
	// ************************************************

	TDS_OUTB(reg, TDS_SAMPLING_START);
	TDS_SC_DEBUG(tds_debug_medium, "done");
}


static TDS_INLINE void
tdsreg_stop_sampling(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_SAMPLING_START);
	TDS_OUTB(reg, TDS_SAMPLING_STOP);
	TDS_SC_DEBUG(tds_debug_medium, "done");
}


static TDS_INLINE u32
tdsreg_get_data(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_DATA);
	u32 val = TDS_INL(reg);
	return val;
}


static TDS_INLINE u8
tdsreg_get_input_status(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_INPUT);
	u8 val = TDS_INB(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_dma_address(struct tds_softc *sc, u32 dstaddr)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_DMA_ADDR_SET);
	TDS_OUTL(reg, dstaddr);
}


static TDS_INLINE void
tdsreg_set_dma_datasize(struct tds_softc *sc, u32 dsize)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_DMA_SIZE_SET);
	TDS_OUTL(reg, dsize);
}


static TDS_INLINE u32
tdsreg_get_dma_datasize(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_DMA_SIZE);
	u32 val = TDS_INL(reg);
	return val;
}


static TDS_INLINE void
tdsreg_start_dma(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_DMA_START);
	TDS_OUTB(reg, TDS_DMA_START);
	TDS_SC_DEBUG(tds_debug_medium, "done");
}

static TDS_INLINE void
tdsreg_stop_dma(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_DMA_START);
	TDS_OUTB(reg, TDS_DMA_STOP);
	TDS_SC_DEBUG(tds_debug_medium, "done");
}


static TDS_INLINE u8
tdsreg_get_dma_status(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_DMA_STATUS);
	u8 val = TDS_INB(reg);
	return val;
}


static TDS_INLINE void
tdsreg_set_dma_mode(struct tds_softc *sc, u8 mode)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_DMA_MODE);
	TDS_OUTB(reg, mode);
}


#if 0  
static TDS_INLINE u8
tdsreg_get_dma_mode(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_DMA_MODE);
	u8 val = TDS_INB(reg);
	return val;
}


static TDS_INLINE u32
tdsreg_get_write_counter(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_WRITE_COUNTER);
	u32 val  = TDS_INL(reg);
	return val;
}



static TDS_INLINE u32
tdsreg_get_read_counter(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_READ_READ_COUNTER);
	u32 val  = TDS_INL(reg);
	return val;
}


static TDS_INLINE void
tdsreg_clear_write_counter(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_CLEAR_WRITE_COUNTER);
	TDS_OUTB(reg, TDS_FIFO_COUNTER_CLEAR);
}


static TDS_INLINE void
tdsreg_clear_read_counter(struct tds_softc *sc)
{
	u32 reg = TDS_IOPORT(TDS_REG_WRITE_CLEAR_READ_COUNTER);
	TDS_OUTB(reg, TDS_FIFO_COUNTER_CLEAR);
}
#endif 


static TDS_INLINE void
tds_wait_next_1pps(struct tds_softc *sc)
{
	u8 saved_intr = sc->intr_flag;
	unsigned long irq_flag;

	spin_lock_irqsave(&sc->lock, irq_flag);
	sc->intr_flag |= TDS_INTR_1PPS;
	tdsreg_set_irq(sc, sc->intr_flag);
	spin_unlock_irqrestore(&sc->lock, irq_flag);
	interruptible_sleep_on(&sc->wait_sync_1pps);
	sc->intr_flag = saved_intr;
	tdsreg_set_irq(sc, sc->intr_flag);
}


static TDS_INLINE void
tds_issue_dma(struct tds_softc *sc)
{
	//int idx = sc->dmabuf_write;
	
	//tdsreg_set_dma_mode(sc, TDS_DMA_MODE_HARD | TDS_DMA_MODE_RELATIVE); //cancel  TDS_DMA_MODE_RELATIVE  M.K.
	tdsreg_set_dma_mode(sc, TDS_DMA_MODE_HARD | TDS_DMA_MODE_REWIND);     //use     TDS_DMA_MODE_REWIND    M.K.                         

	tdsreg_set_dma_datasize(sc, sc->dma_chunk_size / sizeof(u32));
	//tdsreg_set_dma_address(sc, (u32)buf->busaddr); //cancel M.K.
	tdsreg_set_dma_address(sc,sc->dmabuf[TDS_DMA_NSEG-1].busaddr); //append M.K.

	tdsreg_start_dma(sc);
	TDS_SC_DEBUG(tds_debug_medium,
		     "DMA start (%d word) with buffer[%d] busaddr 0x%lx",
		     sc->dma_chunk_size / sizeof(u32), idx, buf->busaddr);
	return;
}


static TDS_INLINE int
tds_check_extension_card(struct tds_softc *sc)
{
	u8 stat;
	int rv;

	stat = tdsreg_get_input_status(sc);
	if (stat & TDS_EXTCARD_EXIST) {
		rv = tds_true;
	}
	else {
		rv = tds_false;
	}
	return rv;
}


static TDS_INLINE void
tds_dma_buffer_init(struct tds_softc *sc)
{
	int i;

	for (i = 0; i < TDS_DMA_NSEG; i ++) {
		tds_dma_buf_t *buf = &sc->dmabuf[i];
		buf->write_offset = 0;
		buf->read_offset = 0;
		buf->flags = 0;
		if (tds_debug > tds_debug_none) {
			memset((void *)buf->virtaddr, 0, sc->dma_buffer_size);
		}
	}
	sc->dmabuf_write = 0;
	sc->dmabuf_read = 0;

	tds_issue_dma(sc);
}


static TDS_INLINE void
tds_calc_dma_chunk_size(struct tds_softc *sc)
{
#if (TDS_DMA_USE_CHUNK == 0)
	/* bps: bits per sec */
	int bps = sc->sampling_freq * sc->sampling_width * sc->sampling_channel;
	/* we use chunk size 32KWord per 1Mbps */
	size_t size = 32 * bps / 1024;
	size_t chunk;

	for (chunk = TDS_DMA_MIN_CHUNK; size > chunk; chunk *= 2)
		continue;
	if (chunk > TDS_DMA_MAX_CHUNK) {
		chunk = TDS_DMA_MAX_CHUNK;
	}
	sc->dma_chunk_size = chunk * sizeof(u32);
#else	/* (TDS_DMA_USE_CHUNK == 0) */
	sc->dma_chunk_size = TDS_DMA_USE_CHUNK * sizeof(u32);
#endif	/* (TDS_DMA_USE_CHUNK == 0) */
	TDS_SC_DEBUG(tds_debug_medium, "DMA chunk_size = 0x%05x(word)",
		     sc->dma_chunk_size / sizeof(u32));
}


static TDS_INLINE int
tds_sampling_condition(struct tds_softc *sc, u32 cond)
{
	int rv = 0;
	int freq = 0;
	int width = 0;
	int channel = 0;

	switch (cond & TDS_MASK_SAMP_FREQ) {
	case TDS_SAMPLING_40KHZ:
		freq = 40 * 1000;
		break;
	case TDS_SAMPLING_100KHZ:
		freq = 100 * 1000;
		break;
	case TDS_SAMPLING_200KHZ:
		freq = 200 * 1000;
		break;
	case TDS_SAMPLING_500KHZ:
		freq = 500 * 1000;
		break;
	case TDS_SAMPLING_1MHZ:
		freq = 1 * 1000000;
		break;
	case TDS_SAMPLING_2MHZ:
		freq = 2 * 1000000;
		break;
	case TDS_SAMPLING_4MHZ:
		freq = 4 * 1000000;
		break;
	case TDS_SAMPLING_8MHZ:
		freq = 8 * 1000000;
		break;
	case TDS_SAMPLING_16MHZ:
		freq = 16 * 1000000;
		break;
	default:
		TDS_SC_PRINT(KERN_ERR, "unsupported freqency");
		sc->error = TDS_ERROR_NOT_SUPPORTED;
		rv = (-ENODEV);
		break;
	}
	switch (cond & TDS_MASK_SAMP_WIDTH) {
	case TDS_SAMPLING_1BIT:
		width = 1;
		break;
	case TDS_SAMPLING_2BIT:
		width = 2;
		break;
	case TDS_SAMPLING_4BIT:
		width = 4;
		break;
	case TDS_SAMPLING_8BIT:
		width = 8;
		break;
	}
	switch (cond & TDS_MASK_SAMP_CH) {
	case TDS_SAMPLING_1CH:
		channel = 1;
		break;
	case TDS_SAMPLING_4CH:
		channel = 4;

		/* it requires extension card, check it out */
		if (tds_check_extension_card(sc) == tds_false) {
			TDS_SC_PRINT(KERN_ERR, "extension card is NOT installed");
			sc->error = TDS_ERROR_EXTCARD_NOT_EXIST;
#ifdef NOT_YET
			rv = (-ENXIO);
#endif
		}
		break;
	}
	if (rv == 0) { /* valid */
		sc->sampling_freq = freq;
		sc->sampling_width = width;
		sc->sampling_channel = channel;
		tds_calc_dma_chunk_size(sc);
		TDS_SC_DEBUG(tds_debug_medium,
			     "sampling condition %dHz/%dbit/%dch",
			     freq, width, channel);
	}
	return rv;
}


static TDS_INLINE void
tds_dma_buf_free(struct tds_softc *sc)
{
	int index;
	for (index = TDS_DMA_NSEG - 1; index >= 0; index--) {
		tds_dma_buf_t *buf = &(sc->dmabuf[index]);
		if (buf->virtaddr != 0) {
			free_pages(buf->virtaddr, buf->order);
			TDS_SC_DEBUG(tds_debug_most, "free_pages (%d)", index);
			buf->virtaddr = 0;
		}
	}
}


static TDS_INLINE int
tds_dma_buf_alloc(struct tds_softc *sc)
{
	int fail_count;
	const int fail_max = 3;
	int index;
	size_t dma_size = sc->dma_buffer_size * 2;	/* for safety */
	unsigned int ord;
	
	ord = get_order(dma_size);
	TDS_SC_DEBUG(tds_debug_most, "get_free_pages (order = %d)", ord);

	for (index = 0; index < TDS_DMA_NSEG; index ++) {
		sc->dmabuf[index].sc = sc;
	}

	fail_count = 0;
	index = 0;
	while ((index < TDS_DMA_NSEG) && (fail_count < fail_max)) {
		tds_dma_buf_t *buf = &(sc->dmabuf[index]);
		if(index==TDS_DMA_NSEG-1) ord=get_order(TDS_DMA_MAX_CHUNK*sizeof(int)*4); //append by M.K.
		else                      ord=get_order(TDS_DMA_MAX_CHUNK*sizeof(int)*2); //append by M.K.
		buf->order = ord;
      
		buf->virtaddr = __get_free_pages(GFP_KERNEL, buf->order);
		if (buf->virtaddr == 0) {
			TDS_SC_PRINT(KERN_ERR,
				     "get_free_pages failure.(%d)", index);
			buf->virtaddr = 0;
			fail_count ++;
			continue;
		}
		TDS_SC_DEBUG(tds_debug_most,
			     "get_free_pages succeed. (%d)", index);
		buf->busaddr = virt_to_bus((void *)buf->virtaddr);
		index ++;
	}
	if (fail_count >= fail_max) {
		TDS_SC_PRINT(KERN_ERR, "get_free_pages too many failure.");
		tds_dma_buf_free(sc);
		return (-ENXIO);
	}

	return (0);
}


static TDS_INLINE int
tds_init(struct tds_softc *sc, int mode)
{
	int rv = 0;

	sc->error = TDS_ERROR_NO_ERROR;
	sc->read_mode = mode;
	sc->max_fifo_overflow = TDS_MAX_FIFO_OVERFLOW_DEFAULT;
	sc->sampling_flag = TDS_SAMPLING_DEFAULT;
	sc->intr_flag = 0;

	sc->is_open = tds_false;
	sc->sampling_done = tds_false;
	sc->pps_sync_done = tds_false;
	sc->time_sync_done = tds_false;
	sc->time_set_done = tds_false;
	sc->sampling_start = tds_false;

	spin_lock_init(&sc->lock);
	sema_init(&sc->sem, 1);

	init_waitqueue_head(&sc->wait_sync_1pps);
	init_waitqueue_head(&sc->wait_sync_data);
	init_waitqueue_head(&sc->wait_buffer_ready);
	init_waitqueue_head(&sc->wait_fifo_underflow);
	init_waitqueue_head(&sc->wait_fifo_overhalf);
	init_waitqueue_head(&sc->wait_fifo_overflow);

	sc->fifo_overflow_count = 0;

	sc->dma_buffer_size = TDS_DMA_SIZE;
	sc->dma_chunk_size = TDS_DMA_MIN_CHUNK * sizeof(u32);	/* fail safe */

	switch (sc->read_mode) {
	case TDS_READ_USE_PIO:
		tds_dma_buf_free(sc);
		break;
	case TDS_READ_USE_DMA:
		rv = tds_dma_buf_alloc(sc);
		break;
	default:
		panic("%s : unknown read mode (%d)",
		      sc->dev_name, sc->read_mode);
		break;
	}

	if (tds_sampling_condition(sc, sc->sampling_flag) == 0) {
		tdsreg_set_sampling(sc, sc->sampling_flag);
	}

	return 0;
}


#if 0
static TDS_INLINE int
tds_get_fifo_count(struct tds_softc *sc)
{
	int wcount = tdsreg_get_write_counter(sc);
	int rcount = tdsreg_get_read_counter(sc);

	if (wcount < rcount) {
		wcount += 0x40000000;
	}
	return (wcount - rcount);
}
#endif


static TDS_INLINE int
tdsio_sampling_start(struct tds_softc *sc)
{
	int rv = 0;
	u8 status;
	unsigned long irq_flag;

	TDS_SC_DEBUG(tds_debug_minimum, "called");

	status = tdsreg_get_status(sc);
	if ((status & TDS_STAT_10MHZ_INPUT) == 0) {
		TDS_SC_PRINT(KERN_ERR, "10MHz input not detected");
		sc->error = TDS_ERROR_10MHZ_NOT_DETECT;
		rv = (-EIO);
	}
	else if (sc->pps_sync_done == tds_false) {
		TDS_SC_PRINT(KERN_ERR, "not synchronized to 1pps yet");
		sc->error = TDS_ERROR_NOT_SYNC_1PPS;
		rv = (-EIO);
	}
	else if ((omit_timeset_check == 0)
		 && ((sc->time_sync_done  | sc->time_set_done) == 0)) {
		TDS_SC_PRINT(KERN_ERR, "not synchronized (or set) time yet");
		sc->error = TDS_ERROR_NOT_SYNC_TIME;
		rv = (-EIO);
	}
	else {
		switch (sc->read_mode) {
		case TDS_READ_USE_PIO:
			spin_lock_irqsave(&sc->lock, irq_flag);
			sc->intr_flag = TDS_INTR_SET_PIO_MODE;
			tdsreg_set_irq(sc, sc->intr_flag);
			sc->sampling_start = tds_true;
			tdsreg_start_sampling(sc);
			spin_unlock_irqrestore(&sc->lock, irq_flag);
			break;
		case TDS_READ_USE_DMA:
			spin_lock_irqsave(&sc->lock, irq_flag);
			tdsreg_clear_fifo(sc); //append M.K.
			tds_dma_buffer_init(sc);
			sc->intr_flag = TDS_INTR_SET_DMA_MODE;
			tdsreg_set_irq(sc, sc->intr_flag);
			sc->sampling_start = tds_true;
			tdsreg_start_sampling(sc);
			spin_unlock_irqrestore(&sc->lock, irq_flag);
			/* actual sampling will start after next 1PPS sync */
			break;
		default:
			panic("%s: unknown read mode (%d)",
			      sc->dev_name, sc->read_mode);
			break;
		}
	}
	return rv;
}



static TDS_INLINE void
tdsio_sampling_stop(struct tds_softc *sc)
{
	unsigned long irq_flag;

	TDS_SC_DEBUG(tds_debug_minimum, "called");

	switch(sc->read_mode) {
	case TDS_READ_USE_PIO:
		spin_lock_irqsave(&sc->lock, irq_flag);
		sc->sampling_start = tds_false;
		tdsreg_stop_sampling(sc);
		spin_unlock_irqrestore(&sc->lock, irq_flag);
		break;
	case TDS_READ_USE_DMA:
		spin_lock_irqsave(&sc->lock, irq_flag);
		tdsreg_stop_dma(sc);
		tdsreg_set_dma_mode(sc, TDS_DMA_MODE_SOFT);
		sc->sampling_start = tds_false;
		tdsreg_stop_sampling(sc);
		/* wait for next 1PPS sync intrrupt before disabling DMA */
		sc->intr_flag |= TDS_INTR_1PPS;
		tdsreg_set_irq(sc, sc->intr_flag);
		spin_unlock_irqrestore(&sc->lock, irq_flag);
		interruptible_sleep_on(&sc->wait_sync_1pps);
		spin_lock_irqsave(&sc->lock, irq_flag);
		sc->intr_flag = 0;
		tdsreg_set_irq(sc, sc->intr_flag);
		spin_unlock_irqrestore(&sc->lock, irq_flag);
		break;
	default:
		panic("%s: unknown read mode (%d)",
		      sc->dev_name, sc->read_mode);
		break;
	}
}


/*
Local Variables:
mode: C
tab-width: 8
c-basic-offset: 8
c-indentation-style: K&R
End:
*/
